@@ -3,14 +3,14 @@
 // (online at: https://github.com/unicode-org/icu4x/blob/main/LICENSE ).
 
 #[cfg(test)]
-#[cfg(feature = "serialize")]
+#[cfg(feature = "datagen")]
 mod test;
 
 use crate::buf::BufferMarker;
 use crate::error::{DataError, DataErrorKind};
-use crate::marker::DataMarker;
+use crate::marker::{DataMarker, ResourceMarker};
 use crate::resource::ResourceKey;
-use crate::resource::ResourcePath;
+use crate::resource::ResourceOptions;
 use crate::yoke::trait_hack::YokeTraitHack;
 use crate::yoke::*;
 
@@ -22,76 +22,20 @@ use core::fmt::Debug;
 use core::marker::PhantomData;
 use icu_locid::LanguageIdentifier;
 
-/// A struct to request a certain piece of data from a data provider.
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Default, Debug, PartialEq, Eq, PartialOrd, Ord)]
+#[non_exhaustive]
+pub struct DataRequestMetadata;
+
+#[derive(Default, Debug, PartialEq, Eq, PartialOrd, Ord)]
+#[allow(clippy::exhaustive_structs)] // this type is stable
 pub struct DataRequest {
-    pub resource_path: ResourcePath,
+    pub options: ResourceOptions,
+    pub metadata: DataRequestMetadata,
 }
 
 impl fmt::Display for DataRequest {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(
-            f,
-            "{}/{}",
-            self.resource_path.key, self.resource_path.options
-        )
-    }
-}
-
-/// Create a [`DataRequest`] to a particular [`ResourceKey`] with default options.
-impl From<ResourceKey> for DataRequest {
-    fn from(key: ResourceKey) -> Self {
-        Self {
-            resource_path: ResourcePath {
-                key,
-                options: Default::default(),
-            },
-        }
-    }
-}
-
-impl DataRequest {
-    /// Returns the [`LanguageIdentifier`] for this [`DataRequest`], or an error if it is not present.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use icu_provider::prelude::*;
-    ///
-    /// const FOO_BAR: ResourceKey = icu_provider::resource_key!("foo/bar@1");
-    ///
-    /// let req_no_langid = DataRequest {
-    ///     resource_path: ResourcePath {
-    ///         key: FOO_BAR,
-    ///         options: ResourceOptions::default(),
-    ///     }
-    /// };
-    ///
-    /// let req_with_langid = DataRequest {
-    ///     resource_path: ResourcePath {
-    ///         key: FOO_BAR,
-    ///         options: ResourceOptions {
-    ///             variant: None,
-    ///             langid: Some(icu_locid_macros::langid!("ar-EG")),
-    ///         },
-    ///     }
-    /// };
-    ///
-    /// assert!(matches!(
-    ///     req_no_langid.try_langid(),
-    ///     Err(DataError { kind: DataErrorKind::NeedsLocale, .. })
-    /// ));
-    /// assert!(matches!(
-    ///     req_with_langid.try_langid(),
-    ///     Ok(_)
-    /// ));
-    /// ```
-    pub fn try_langid(&self) -> Result<&LanguageIdentifier, DataError> {
-        self.resource_path
-            .options
-            .langid
-            .as_ref()
-            .ok_or_else(|| DataErrorKind::NeedsLocale.with_req(self))
+        fmt::Display::fmt(&self.options, f)
     }
 }
 
@@ -105,7 +49,7 @@ pub struct DataResponseMetadata {
     pub buffer_format: Option<crate::buf::BufferFormat>,
 }
 
-/// A container for data payloads returned from a [`DataProvider`].
+/// A container for data payloads returned from a data provider.
 ///
 /// [`DataPayload`] is built on top of the [`yoke`] framework, which allows for cheap, zero-copy
 /// operations on data via the use of self-references. A [`DataPayload`] may be backed by one of
@@ -477,10 +421,7 @@ where
     ///
     /// Same example as above, but this time, do not move out of `p1`:
     ///
-    /// ***[#1061](https://github.com/unicode-org/icu4x/issues/1061): The following example
-    /// requires Rust 1.57.***
-    ///
-    /// ```ignore
+    /// ```
     /// // Same imports and definitions as above
     /// # use icu_provider::hello_world::*;
     /// # use icu_provider::prelude::*;
@@ -526,10 +467,7 @@ where
     ///
     /// Capture a string from the context and append it to the message:
     ///
-    /// ***[#1061](https://github.com/unicode-org/icu4x/issues/1061): The following example
-    /// requires Rust 1.57.***
-    ///
-    /// ```ignore
+    /// ```
     /// // Same imports and definitions as above
     /// # use icu_provider::hello_world::*;
     /// # use icu_provider::prelude::*;
@@ -554,34 +492,6 @@ where
     ///
     /// assert_eq!("Hello WorldExtra", p2.get());
     /// ```
-    ///
-    /// Prior to Rust 1.57, pass the capture by value instead of by reference:
-    ///
-    /// ```
-    /// // Same imports and definitions as above
-    /// # use icu_provider::hello_world::*;
-    /// # use icu_provider::prelude::*;
-    /// # use std::borrow::Cow;
-    /// # struct HelloWorldV1MessageMarker;
-    /// # impl DataMarker for HelloWorldV1MessageMarker {
-    /// #     type Yokeable = Cow<'static, str>;
-    /// # }
-    ///
-    /// let p1: DataPayload<HelloWorldV1Marker> = DataPayload::from_owned(HelloWorldV1 {
-    ///     message: Cow::Borrowed("Hello World")
-    /// });
-    ///
-    /// assert_eq!("Hello World", p1.get().message);
-    ///
-    /// let p2: DataPayload<HelloWorldV1MessageMarker> = p1.map_project_with_capture(
-    ///     "Extra".to_string(),
-    ///     |mut obj, capture, _| {
-    ///         obj.message.to_mut().push_str(&capture);
-    ///         obj.message
-    ///     });
-    ///
-    /// assert_eq!("Hello WorldExtra", p2.get());
-    /// ```
     #[allow(clippy::type_complexity)]
     pub fn map_project_with_capture<M2, T>(
         self,
@@ -607,10 +517,7 @@ where
     ///
     /// Same example as above, but this time, do not move out of `p1`:
     ///
-    /// ***[#1061](https://github.com/unicode-org/icu4x/issues/1061): The following example
-    /// requires Rust 1.57.***
-    ///
-    /// ```ignore
+    /// ```
     /// // Same imports and definitions as above
     /// # use icu_provider::hello_world::*;
     /// # use icu_provider::prelude::*;
@@ -663,10 +570,7 @@ where
     ///
     /// Same example as above, but bubble up an error:
     ///
-    /// ***[#1061](https://github.com/unicode-org/icu4x/issues/1061): The following example
-    /// requires Rust 1.57.***
-    ///
-    /// ```ignore
+    /// ```
     /// // Same imports and definitions as above
     /// # use icu_provider::hello_world::*;
     /// # use icu_provider::prelude::*;
@@ -695,38 +599,6 @@ where
     /// assert_eq!("Hello WorldExtra", p2.get());
     /// # Ok::<(), &'static str>(())
     /// ```
-    ///
-    /// Prior to Rust 1.57, pass the capture by value instead of by reference:
-    ///
-    /// ```
-    /// // Same imports and definitions as above
-    /// # use icu_provider::hello_world::*;
-    /// # use icu_provider::prelude::*;
-    /// # use std::borrow::Cow;
-    /// # struct HelloWorldV1MessageMarker;
-    /// # impl DataMarker for HelloWorldV1MessageMarker {
-    /// #     type Yokeable = Cow<'static, str>;
-    /// # }
-    ///
-    /// let p1: DataPayload<HelloWorldV1Marker> = DataPayload::from_owned(HelloWorldV1 {
-    ///     message: Cow::Borrowed("Hello World")
-    /// });
-    ///
-    /// assert_eq!("Hello World", p1.get().message);
-    ///
-    /// let p2: DataPayload<HelloWorldV1MessageMarker> = p1.try_map_project_with_capture(
-    ///     "Extra".to_string(),
-    ///     |mut obj, capture, _| {
-    ///         if obj.message.is_empty() {
-    ///             return Err(())
-    ///         }
-    ///         obj.message.to_mut().push_str(&capture);
-    ///         Ok(obj.message)
-    ///     })?;
-    ///
-    /// assert_eq!("Hello WorldExtra", p2.get());
-    /// # Ok::<(), ()>(())
-    /// ```
     #[allow(clippy::type_complexity)]
     pub fn try_map_project_with_capture<M2, T, E>(
         self,
@@ -752,10 +624,7 @@ where
     ///
     /// Same example as above, but bubble up an error:
     ///
-    /// ***[#1061](https://github.com/unicode-org/icu4x/issues/1061): The following example
-    /// requires Rust 1.57.***
-    ///
-    /// ```ignore
+    /// ```
     /// // Same imports and definitions as above
     /// # use icu_provider::hello_world::*;
     /// # use icu_provider::prelude::*;
@@ -804,6 +673,14 @@ where
             yoke: self.yoke.try_project_cloned_with_capture(capture, f)?,
         })
     }
+
+    #[inline]
+    pub fn cast<M2>(self) -> DataPayload<M2>
+    where
+        M2: DataMarker<Yokeable = M::Yokeable>,
+    {
+        DataPayload { yoke: self.yoke }
+    }
 }
 
 impl DataPayload<BufferMarker> {
@@ -840,6 +717,7 @@ where
 }
 
 /// A response object containing an object as payload and metadata about it.
+#[allow(clippy::exhaustive_structs)] // this type is stable
 pub struct DataResponse<M>
 where
     M: DataMarker,
@@ -939,14 +817,13 @@ fn test_debug() {
     assert_eq!("DataResponse { metadata: DataResponseMetadata { data_langid: None, buffer_format: None }, payload: Some(HelloWorldV1 { message: \"foo\" }) }", format!("{:?}", resp));
 }
 
-/// A generic data provider that loads a payload of a specific type.
+/// A data provider that loads data for a specific data type.
 ///
-/// See examples on some of the concrete implementations:
+/// Unlike [`ResourceProvider`], there may be multiple keys corresponding to the same data type.
+/// This is often the case when returning `dyn` trait objects such as [`SerializeMarker`].
 ///
-/// - [`HelloWorldProvider`](crate::hello_world::HelloWorldProvider)
-/// - [`AnyPayloadProvider`](crate::struct_provider::AnyPayloadProvider)
-/// - [`InvariantDataProvider`](crate::inv::InvariantDataProvider)
-pub trait DataProvider<M>
+/// [`SerializeMarker`]: crate::serde::SerializeMarker
+pub trait DynProvider<M>
 where
     M: DataMarker,
 {
@@ -954,5 +831,35 @@ where
     ///
     /// Returns [`Ok`] if the request successfully loaded data. If data failed to load, returns an
     /// Error with more information.
-    fn load_payload(&self, req: &DataRequest) -> Result<DataResponse<M>, DataError>;
+    fn load_payload(
+        &self,
+        key: ResourceKey,
+        req: &DataRequest,
+    ) -> Result<DataResponse<M>, DataError>;
+}
+
+/// A data provider that loads data for a specific [`ResourceKey`].
+pub trait ResourceProvider<M>
+where
+    M: ResourceMarker,
+{
+    /// Query the provider for data, returning the result.
+    ///
+    /// Returns [`Ok`] if the request successfully loaded data. If data failed to load, returns an
+    /// Error with more information.
+    fn load_resource(&self, req: &DataRequest) -> Result<DataResponse<M>, DataError>;
+}
+
+impl<M, P> DynProvider<M> for alloc::boxed::Box<P>
+where
+    M: DataMarker,
+    P: DynProvider<M> + ?Sized,
+{
+    fn load_payload(
+        &self,
+        key: ResourceKey,
+        req: &DataRequest,
+    ) -> Result<DataResponse<M>, DataError> {
+        (**self).load_payload(key, req)
+    }
 }
