@@ -8,9 +8,11 @@ use core::fmt::Debug;
 use serde::{Deserialize, Serialize};
 
 use super::*;
+use crate::datagen::IterableResourceProvider;
 use crate::hello_world::*;
 use crate::prelude::*;
 use crate::yoke;
+use crate::zerofrom;
 
 // This file tests DataProvider borrow semantics with a dummy data provider based on a
 // JSON string. It also exercises most of the data provider code paths.
@@ -19,9 +21,11 @@ use crate::yoke;
 const HELLO_ALT_KEY: ResourceKey = crate::resource_key!("core/helloalt@1");
 
 /// A data struct serialization-compatible with HelloWorldV1 used for testing mismatched types
-#[derive(Serialize, Deserialize, Debug, Clone, Default, PartialEq, Yokeable, ZeroCopyFrom)]
-#[yoke(cloning_zcf)]
+#[derive(
+    Serialize, Deserialize, Debug, Clone, Default, PartialEq, Yokeable, zerofrom::ZeroFrom,
+)]
 struct HelloAlt {
+    #[zerofrom(clone)]
     message: String,
 }
 
@@ -32,6 +36,10 @@ impl DataMarker for HelloAltMarker {
     type Yokeable = HelloAlt;
 }
 
+impl ResourceMarker for HelloAltMarker {
+    const KEY: ResourceKey = HELLO_ALT_KEY;
+}
+
 #[derive(Serialize, Deserialize, Debug, Clone, Default, PartialEq)]
 struct HelloCombined<'data> {
     #[serde(borrow)]
@@ -47,12 +55,11 @@ struct DataWarehouse {
     hello_alt: HelloAlt,
 }
 
-impl DataProvider<HelloWorldV1Marker> for DataWarehouse {
-    fn load_payload(
+impl ResourceProvider<HelloWorldV1Marker> for DataWarehouse {
+    fn load_resource(
         &self,
-        req: &DataRequest,
+        _: &DataRequest,
     ) -> Result<DataResponse<HelloWorldV1Marker>, DataError> {
-        req.resource_path.key.match_key(key::HELLO_WORLD_V1)?;
         Ok(DataResponse {
             metadata: DataResponseMetadata::default(),
             payload: Some(DataPayload::from_owned(self.hello_v1.clone())),
@@ -60,9 +67,13 @@ impl DataProvider<HelloWorldV1Marker> for DataWarehouse {
     }
 }
 
-crate::impl_dyn_provider!(DataWarehouse, {
-    key::HELLO_WORLD_V1 => HelloWorldV1Marker,
-}, ANY);
+impl IterableResourceProvider<HelloWorldV1Marker> for DataWarehouse {
+    fn supported_options(&self) -> Result<Box<dyn Iterator<Item = ResourceOptions>>, DataError> {
+        Ok(Box::new(core::iter::once(ResourceOptions::default())))
+    }
+}
+
+crate::impl_dyn_provider!(DataWarehouse, [HelloWorldV1Marker,], ANY);
 
 /// A DataProvider that supports both key::HELLO_WORLD_V1 and HELLO_ALT.
 #[derive(Debug)]
@@ -76,12 +87,11 @@ impl From<DataWarehouse> for DataProvider2 {
     }
 }
 
-impl DataProvider<HelloWorldV1Marker> for DataProvider2 {
-    fn load_payload(
+impl ResourceProvider<HelloWorldV1Marker> for DataProvider2 {
+    fn load_resource(
         &self,
-        req: &DataRequest,
+        _: &DataRequest,
     ) -> Result<DataResponse<HelloWorldV1Marker>, DataError> {
-        req.resource_path.key.match_key(key::HELLO_WORLD_V1)?;
         Ok(DataResponse {
             metadata: DataResponseMetadata::default(),
             payload: Some(DataPayload::from_owned(self.data.hello_v1.clone())),
@@ -89,9 +99,14 @@ impl DataProvider<HelloWorldV1Marker> for DataProvider2 {
     }
 }
 
-impl DataProvider<HelloAltMarker> for DataProvider2 {
-    fn load_payload(&self, req: &DataRequest) -> Result<DataResponse<HelloAltMarker>, DataError> {
-        req.resource_path.key.match_key(HELLO_ALT_KEY)?;
+impl IterableResourceProvider<HelloWorldV1Marker> for DataProvider2 {
+    fn supported_options(&self) -> Result<Box<dyn Iterator<Item = ResourceOptions>>, DataError> {
+        Ok(Box::new(core::iter::once(ResourceOptions::default())))
+    }
+}
+
+impl ResourceProvider<HelloAltMarker> for DataProvider2 {
+    fn load_resource(&self, _: &DataRequest) -> Result<DataResponse<HelloAltMarker>, DataError> {
         Ok(DataResponse {
             metadata: DataResponseMetadata::default(),
             payload: Some(DataPayload::from_owned(self.data.hello_alt.clone())),
@@ -99,10 +114,13 @@ impl DataProvider<HelloAltMarker> for DataProvider2 {
     }
 }
 
-crate::impl_dyn_provider!(DataProvider2, {
-    key::HELLO_WORLD_V1 => HelloWorldV1Marker,
-    HELLO_ALT_KEY => HelloAltMarker,
-}, ANY);
+impl IterableResourceProvider<HelloAltMarker> for DataProvider2 {
+    fn supported_options(&self) -> Result<Box<dyn Iterator<Item = ResourceOptions>>, DataError> {
+        Ok(Box::new(core::iter::once(ResourceOptions::default())))
+    }
+}
+
+crate::impl_dyn_provider!(DataProvider2, [HelloWorldV1Marker, HelloAltMarker,], ANY);
 
 #[allow(clippy::redundant_static_lifetimes)]
 const DATA: &'static str = r#"{
@@ -123,34 +141,16 @@ fn get_warehouse(data: &'static str) -> DataWarehouse {
     }
 }
 
-fn get_payload_v1<P: DataProvider<HelloWorldV1Marker> + ?Sized>(
+fn get_payload_v1<P: ResourceProvider<HelloWorldV1Marker> + ?Sized>(
     provider: &P,
 ) -> Result<DataPayload<HelloWorldV1Marker>, DataError> {
-    provider.load_payload(&get_request_v1())?.take_payload()
+    provider.load_resource(&Default::default())?.take_payload()
 }
 
-fn get_payload_alt<P: DataProvider<HelloAltMarker> + ?Sized>(
-    d: &P,
+fn get_payload_alt<P: ResourceProvider<HelloAltMarker> + ?Sized>(
+    provider: &P,
 ) -> Result<DataPayload<HelloAltMarker>, DataError> {
-    d.load_payload(&get_request_alt())?.take_payload()
-}
-
-fn get_request_v1() -> DataRequest {
-    DataRequest {
-        resource_path: ResourcePath {
-            key: key::HELLO_WORLD_V1,
-            options: Default::default(),
-        },
-    }
-}
-
-fn get_request_alt() -> DataRequest {
-    DataRequest {
-        resource_path: ResourcePath {
-            key: HELLO_ALT_KEY,
-            options: Default::default(),
-        },
-    }
+    provider.load_resource(&Default::default())?.take_payload()
 }
 
 #[test]
@@ -180,7 +180,8 @@ fn test_warehouse_owned_dyn_erased() {
 #[test]
 fn test_warehouse_owned_dyn_generic() {
     let warehouse = get_warehouse(DATA);
-    let hello_data = get_payload_v1(&warehouse as &dyn DataProvider<HelloWorldV1Marker>).unwrap();
+    let hello_data =
+        get_payload_v1(&warehouse as &dyn ResourceProvider<HelloWorldV1Marker>).unwrap();
     assert!(matches!(
         hello_data.get(),
         HelloWorldV1 {
@@ -240,7 +241,8 @@ fn test_provider2_dyn_erased_alt() {
 fn test_provider2_dyn_generic() {
     let warehouse = get_warehouse(DATA);
     let provider = DataProvider2::from(warehouse);
-    let hello_data = get_payload_v1(&provider as &dyn DataProvider<HelloWorldV1Marker>).unwrap();
+    let hello_data =
+        get_payload_v1(&provider as &dyn ResourceProvider<HelloWorldV1Marker>).unwrap();
     assert!(matches!(
         hello_data.get(),
         HelloWorldV1 {
@@ -253,7 +255,7 @@ fn test_provider2_dyn_generic() {
 fn test_provider2_dyn_generic_alt() {
     let warehouse = get_warehouse(DATA);
     let provider = DataProvider2::from(warehouse);
-    let hello_data = get_payload_alt(&provider as &dyn DataProvider<HelloAltMarker>).unwrap();
+    let hello_data = get_payload_alt(&provider as &dyn ResourceProvider<HelloAltMarker>).unwrap();
     assert!(matches!(hello_data.get(), HelloAlt { .. }));
 }
 
@@ -262,10 +264,13 @@ fn test_mismatched_types() {
     let warehouse = get_warehouse(DATA);
     let provider = DataProvider2::from(warehouse);
     // Request is for v2, but type argument is for v1
-    let response: Result<DataResponse<HelloWorldV1Marker>, DataError> =
-        AnyProvider::load_any(&provider.as_any_provider(), &get_request_alt())
-            .unwrap()
-            .downcast();
+    let response: Result<DataResponse<HelloWorldV1Marker>, DataError> = AnyProvider::load_any(
+        &provider.as_any_provider(),
+        HELLO_ALT_KEY,
+        &Default::default(),
+    )
+    .unwrap()
+    .downcast();
     assert!(matches!(
         response,
         Err(DataError {
@@ -277,15 +282,15 @@ fn test_mismatched_types() {
 
 fn check_v1_v2<P>(d: &P)
 where
-    P: DataProvider<HelloWorldV1Marker> + DataProvider<HelloAltMarker> + ?Sized,
+    P: ResourceProvider<HelloWorldV1Marker> + ResourceProvider<HelloAltMarker> + ?Sized,
 {
     let v1: DataPayload<HelloWorldV1Marker> = d
-        .load_payload(&get_request_v1())
+        .load_resource(&Default::default())
         .unwrap()
         .take_payload()
         .unwrap();
     let v2: DataPayload<HelloAltMarker> = d
-        .load_payload(&get_request_alt())
+        .load_resource(&Default::default())
         .unwrap()
         .take_payload()
         .unwrap();
