@@ -9,6 +9,7 @@ mod impls;
 pub use impls::*;
 
 use crate::yoke::Yokeable;
+use crate::ResourceKey;
 
 /// Trait marker for data structs. All types delivered by the data provider must be associated with
 /// something implementing this trait.
@@ -19,7 +20,7 @@ use crate::yoke::Yokeable;
 /// for the data struct:
 ///
 /// - `impl<'a> Yokeable<'a>` (required)
-/// - `impl ZeroCopyFrom<Self>`
+/// - `impl ZeroFrom<Self>`
 ///
 /// See also some common pre-made DataMarker impls in this module.
 ///
@@ -30,10 +31,11 @@ use crate::yoke::Yokeable;
 /// ```
 /// use icu_provider::prelude::*;
 /// use icu_provider::yoke::*;
+/// use icu_provider::zerofrom::*;
 /// use std::borrow::Cow;
 /// use std::rc::Rc;
 ///
-/// #[derive(Yokeable, ZeroCopyFrom)]
+/// #[derive(Yokeable, ZeroFrom)]
 /// struct MyDataStruct<'data> {
 ///     message: Cow<'data, str>,
 /// }
@@ -56,3 +58,7 @@ pub trait DataMarker {
     /// data struct.
     type Yokeable: for<'a> Yokeable<'a>;
 }
+
+pub trait ResourceMarker: DataMarker {
+    const KEY: ResourceKey;
+}
