@@ -8,53 +8,15 @@ use alloc::borrow::Cow;
 
 use crate::error::{DataError, DataErrorKind};
 use crate::helpers;
+use core::cmp::Ordering;
 use core::default::Default;
 use core::fmt;
 use core::fmt::Write;
-use icu_locid::LanguageIdentifier;
+use icu_locid::extensions::unicode as unicode_ext;
+use icu_locid::subtags::{Language, Region, Script};
+use icu_locid::{LanguageIdentifier, Locale};
 use writeable::{LengthHint, Writeable};
-
-/// A compact hash of a [`ResourceKey`]. Useful for keys in maps.
-#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Copy, Clone, Hash)]
-#[repr(transparent)]
-pub struct ResourceKeyHash([u8; 4]);
-
-impl ResourceKeyHash {
-    const fn compute_from_str(path: &str) -> Self {
-        Self(helpers::fxhash_32(path.as_bytes()).to_le_bytes())
-    }
-}
-
-/// Used for loading data from an ICU4X data provider.
-///
-/// A resource key is tightly coupled with the code that uses it to load data at runtime.
-/// Executables can be searched for `ResourceKey` instances to produce optimized data files.
-/// Therefore, users should not generally create ResourceKey instances; they should instead use
-/// the ones exported by a component.
-///
-/// `ResourceKey`s are created with the [`resource_key!`] macro:
-///
-/// ```
-/// # use icu_provider::prelude::ResourceKey;
-/// const K: ResourceKey = icu_provider::resource_key!("foo/bar@1");
-/// ```
-///
-/// The human-readable path string ends with `@` followed by one or more digits (the version
-/// number). Paths do not contain characters other than ASCII letters and digits, `_`, `/`, `=`.
-///
-/// Invalid paths are compile-time errors (as [`resource_key!`] uses `const`).
-///
-/// ```compile_fail,E0080
-/// # use icu_provider::prelude::ResourceKey;
-/// const K: ResourceKey = icu_provider::resource_key!("foo/../bar@1");
-/// ```
-#[derive(PartialEq, Eq, Copy, Clone)]
-pub struct ResourceKey {
-    // This string literal is wrapped in leading_tag!() and trailing_tag!() to make it detectable
-    // in a compiled binary.
-    path: &'static str,
-    hash: ResourceKeyHash,
-}
+use zerovec::ule::*;
 
 #[doc(hidden)]
 #[macro_export]
@@ -84,15 +46,85 @@ macro_rules! tagged {
     };
 }
 
+/// A compact hash of a [`ResourceKey`]. Useful for keys in maps.
+#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Copy, Clone, Hash, ULE)]
+#[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
+#[repr(transparent)]
+pub struct ResourceKeyHash([u8; 4]);
+
+impl ResourceKeyHash {
+    const fn compute_from_str(path: &str) -> Self {
+        Self(
+            helpers::fxhash_32(path.as_bytes(), leading_tag!().len(), trailing_tag!().len())
+                .to_le_bytes(),
+        )
+    }
+}
+
+impl<'a> zerovec::maps::ZeroMapKV<'a> for ResourceKeyHash {
+    type Container = zerovec::ZeroVec<'a, ResourceKeyHash>;
+    type GetType = <ResourceKeyHash as AsULE>::ULE;
+    type OwnedType = ResourceKeyHash;
+}
+
+impl AsULE for ResourceKeyHash {
+    type ULE = Self;
+    #[inline]
+    fn to_unaligned(self) -> Self::ULE {
+        self
+    }
+    #[inline]
+    fn from_unaligned(unaligned: Self::ULE) -> Self {
+        unaligned
+    }
+}
+
+// Safe since the ULE type is `self`.
+unsafe impl EqULE for ResourceKeyHash {}
+
+/// Used for loading data from an ICU4X data provider.
+///
+/// A resource key is tightly coupled with the code that uses it to load data at runtime.
+/// Executables can be searched for `ResourceKey` instances to produce optimized data files.
+/// Therefore, users should not generally create ResourceKey instances; they should instead use
+/// the ones exported by a component.
+///
+/// `ResourceKey`s are created with the [`resource_key!`] macro:
+///
+/// ```
+/// # use icu_provider::prelude::ResourceKey;
+/// const K: ResourceKey = icu_provider::resource_key!("foo/bar@1");
+/// ```
+///
+/// The human-readable path string ends with `@` followed by one or more digits (the version
+/// number). Paths do not contain characters other than ASCII letters and digits, `_`, `/`.
+///
+/// Invalid paths are compile-time errors (as [`resource_key!`] uses `const`).
+///
+/// ```compile_fail,E0080
+/// # use icu_provider::prelude::ResourceKey;
+/// const K: ResourceKey = icu_provider::resource_key!("foo/../bar@1");
+/// ```
+#[derive(PartialEq, Eq, Copy, Clone, PartialOrd, Ord)]
+pub struct ResourceKey {
+    // This string literal is wrapped in leading_tag!() and trailing_tag!() to make it detectable
+    // in a compiled binary.
+    path: &'static str,
+    hash: ResourceKeyHash,
+}
+
+#[cfg(test)]
+static_assertions::const_assert_eq!(24, core::mem::size_of::<ResourceKey>());
+
 impl ResourceKey {
     /// Gets a human-readable representation of a [`ResourceKey`].
     ///
     /// The human-readable path string ends with `@` followed by one or more digits (the version
-    /// number). Paths do not contain characters other than ASCII letters and digits, `_`, `/`, `=`.
+    /// number). Paths do not contain characters other than ASCII letters and digits, `_`, `/`.
     ///
     /// Useful for reading and writing data to a file system.
     #[inline]
-    pub fn get_path(&self) -> &str {
+    pub fn get_path(&self) -> &'static str {
         // This becomes const with `const_ptr_offset` and `const_slice_from_raw_parts`.
         unsafe {
             // Safe due to invariant that self.path is tagged correctly
@@ -115,6 +147,8 @@ impl ResourceKey {
 
     #[doc(hidden)]
     // Error is a str of the expected character class and the index where it wasn't encountered
+    // The indexing operations in this function have been reviewed in detail and won't panic.
+    #[allow(clippy::indexing_slicing)]
     pub const fn construct_internal(path: &'static str) -> Result<Self, (&'static str, usize)> {
         if path.len() < leading_tag!().len() + trailing_tag!().len() {
             return Err(("tag", 0));
@@ -139,7 +173,7 @@ impl ResourceKey {
             i += 1;
         }
 
-        // Regex: [a-zA-Z0-9=_][a-zA-Z0-9=_/]*@[0-9]+
+        // Regex: [a-zA-Z0-9_][a-zA-Z0-9_/]*@[0-9]+
         enum State {
             Empty,
             Body,
@@ -156,7 +190,7 @@ impl ResourceKey {
                 None
             };
             state = match (state, byte) {
-                (Empty | Body, Some(b'a'..=b'z' | b'A'..=b'Z' | b'0'..=b'9' | b'_' | b'=')) => Body,
+                (Empty | Body, Some(b'a'..=b'z' | b'A'..=b'Z' | b'0'..=b'9' | b'_')) => Body,
                 (Body, Some(b'/')) => Body,
                 (Body, Some(b'@')) => At,
                 (At | Version, Some(b'0'..=b'9')) => Version,
@@ -168,8 +202,8 @@ impl ResourceKey {
                     })
                 }
 
-                (Empty, _) => return Err(("[a-zA-Z0-9=_]", i)),
-                (Body, _) => return Err(("[a-zA-z0-9=_/@]", i)),
+                (Empty, _) => return Err(("[a-zA-Z0-9_]", i)),
+                (Body, _) => return Err(("[a-zA-z0-9_/@]", i)),
                 (At | Version, _) => return Err(("[0-9]", i)),
             };
             i += 1;
@@ -230,7 +264,7 @@ fn test_path_syntax() {
     // No version:
     assert_eq!(
         ResourceKey::construct_internal(tagged!("hello/world")),
-        Err(("[a-zA-z0-9=_/@]", 25))
+        Err(("[a-zA-z0-9_/@]", 25))
     );
 
     assert_eq!(
@@ -249,7 +283,7 @@ fn test_path_syntax() {
     // Invalid characters:
     assert_eq!(
         ResourceKey::construct_internal(tagged!("你好/世界@1")),
-        Err(("[a-zA-Z0-9=_]", 14))
+        Err(("[a-zA-Z0-9_]", 14))
     );
 
     // Invalid tag:
@@ -275,6 +309,8 @@ macro_rules! resource_key {
         const RESOURCE_KEY_MACRO_CONST: $crate::ResourceKey = {
             match $crate::ResourceKey::construct_internal($crate::tagged!($path)) {
                 Ok(v) => v,
+                #[allow(clippy::panic)]
+                // TODO(#1668) Clippy exceptions need docs or fixing.
                 Err(_) => panic!(concat!("Invalid resource key: ", $path)),
                 // TODO Once formatting is const:
                 // Err((expected, index)) => panic!(
@@ -313,20 +349,18 @@ impl Writeable for ResourceKey {
         self.get_path().write_len()
     }
 
-    fn writeable_to_string(&self) -> Cow<str> {
+    fn write_to_string(&self) -> Cow<str> {
         Cow::Borrowed(self.get_path())
     }
 }
 
-/// A variant and language identifier, used for requesting data from a
-/// [`DataProvider`](crate::DataProvider).
+/// A variant and language identifier, used for requesting data from a data provider.
 ///
 /// The fields in a [`ResourceOptions`] are not generally known until runtime.
-#[derive(PartialEq, Clone, Default)]
+#[derive(PartialEq, Clone, Default, PartialOrd, Eq, Ord, Hash)]
 pub struct ResourceOptions {
-    // TODO: Consider making multiple variant fields.
-    pub variant: Option<Cow<'static, str>>,
-    pub langid: Option<LanguageIdentifier>,
+    langid: LanguageIdentifier,
+    keywords: unicode_ext::Keywords,
 }
 
 impl fmt::Debug for ResourceOptions {
@@ -343,88 +377,203 @@ impl fmt::Display for ResourceOptions {
 
 impl Writeable for ResourceOptions {
     fn write_to<W: core::fmt::Write + ?Sized>(&self, sink: &mut W) -> core::fmt::Result {
-        let mut initial = true;
-        if let Some(variant) = &self.variant {
-            variant.write_to(sink)?;
-            initial = false;
-        }
-        if let Some(langid) = &self.langid {
-            if !initial {
-                sink.write_char('/')?;
-            }
-            langid.write_to(sink)?;
+        self.langid.write_to(sink)?;
+        if !self.keywords.is_empty() {
+            sink.write_str("-u-")?;
+            self.keywords.write_to(sink)?;
         }
         Ok(())
     }
 
     fn write_len(&self) -> LengthHint {
-        let mut length_hint = LengthHint::exact(0);
-        let mut initial = true;
-        if let Some(variant) = &self.variant {
-            length_hint += variant.write_len();
-            initial = false;
-        }
-        if let Some(langid) = &self.langid {
-            if !initial {
-                length_hint += 1;
+        self.langid.write_len()
+            + if !self.keywords.is_empty() {
+                self.keywords.write_len() + 3
+            } else {
+                LengthHint::exact(0)
             }
-            length_hint += langid.write_len();
-        }
-        length_hint
     }
 }
 
 impl From<LanguageIdentifier> for ResourceOptions {
-    /// Create a ResourceOptions with the given language identifier and an empty variant field.
     fn from(langid: LanguageIdentifier) -> Self {
         Self {
-            langid: Some(langid),
-            variant: None,
+            langid,
+            keywords: unicode_ext::Keywords::new(),
+        }
+    }
+}
+
+impl From<Locale> for ResourceOptions {
+    fn from(locale: Locale) -> Self {
+        // TODO(#1109): Implement proper vertical fallback
+        Self {
+            langid: locale.id,
+            keywords: locale.extensions.unicode.keywords,
+        }
+    }
+}
+
+impl From<&Locale> for ResourceOptions {
+    fn from(locale: &Locale) -> Self {
+        // TODO(#1109): Implement proper vertical fallback
+        Self {
+            langid: locale.id.clone(),
+            keywords: locale.extensions.unicode.keywords.clone(),
         }
     }
 }
 
 impl ResourceOptions {
+    pub fn cmp_bytes(&self, other: &[u8]) -> Ordering {
+        if self.keywords.is_empty() {
+            self.langid.cmp_bytes(other)
+        } else {
+            // TODO: Avoid the allocation
+            self.write_to_string().as_bytes().cmp(other)
+        }
+    }
+}
+
+impl ResourceOptions {
+    /// TODO(#1109): Delete this function and use vertical fallback instead
+    pub fn temp_for_region(region: Option<Region>) -> Self {
+        Self {
+            langid: LanguageIdentifier::from(region),
+            keywords: unicode_ext::Keywords::new(),
+        }
+    }
+
     /// Returns whether this [`ResourceOptions`] has all empty fields (no components).
     pub fn is_empty(&self) -> bool {
         self == &Self::default()
     }
-}
 
-#[derive(Clone, PartialEq)]
-pub struct ResourcePath {
-    pub key: ResourceKey,
-    pub options: ResourceOptions,
-}
-
-impl fmt::Debug for ResourcePath {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "ResourcePath{{{}}}", self)
-    }
-}
-
-impl fmt::Display for ResourcePath {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        writeable::Writeable::write_to(self, f)
-    }
-}
-
-impl writeable::Writeable for ResourcePath {
-    fn write_to<W: core::fmt::Write + ?Sized>(&self, sink: &mut W) -> core::fmt::Result {
-        writeable::Writeable::write_to(&self.key, sink)?;
-        if !self.options.is_empty() {
-            sink.write_char('/')?;
-            writeable::Writeable::write_to(&self.options, sink)?;
-        }
-        Ok(())
+    /// Returns whether the [`LanguageIdentifier`] associated with this request is `und`.
+    ///
+    /// Note that this only checks the language identifier; extension keywords may also be set.
+    /// To check the entire `ResourceOptions`, use [`ResourceOptions::is_empty()`].
+    pub fn is_langid_und(&self) -> bool {
+        self.langid == LanguageIdentifier::UND
     }
 
-    fn write_len(&self) -> writeable::LengthHint {
-        let mut result = writeable::Writeable::write_len(&self.key);
-        if !self.options.is_empty() {
-            result += writeable::Writeable::write_len(&self.options) + 1;
-        }
-        result
+    /// Gets the [`LanguageIdentifier`] for this [`ResourceOptions`].
+    ///
+    /// This may allocate memory if there are variant subtags. If you need only the language,
+    /// script, and/or region subtag, use the specific getters for those subtags:
+    ///
+    /// - [`ResourceOptions::language()`]
+    /// - [`ResourceOptions::script()`]
+    /// - [`ResourceOptions::region()`]
+    ///
+    /// If you have ownership over the `ResourceOptions`, use [`ResourceOptions::into_locale()`]
+    /// and then access the `id` field.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use icu_provider::prelude::*;
+    /// use icu_locid::langid;
+    ///
+    /// const FOO_BAR: ResourceKey = icu_provider::resource_key!("foo/bar@1");
+    ///
+    /// let req_no_langid = DataRequest {
+    ///     options: ResourceOptions::default(),
+    ///     metadata: Default::default(),
+    /// };
+    ///
+    /// let req_with_langid = DataRequest {
+    ///     options: langid!("ar-EG").into(),
+    ///     metadata: Default::default(),
+    /// };
+    ///
+    /// assert_eq!(req_no_langid.options.get_langid(), langid!("und"));
+    /// assert_eq!(req_with_langid.options.get_langid(), langid!("ar-EG"));
+    /// ```
+    pub fn get_langid(&self) -> LanguageIdentifier {
+        self.langid.clone()
+    }
+
+    /// Converts this [`ResourceOptions`] into a [`Locale`].
+    ///
+    /// See also [`ResourceOptions::get_langid()`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use icu_provider::prelude::*;
+    /// use icu_locid::{langid, language, region, Locale};
+    ///
+    /// let locale: Locale = "it-IT-u-ca-coptic".parse()
+    ///     .expect("Valid BCP-47");
+    /// let options: ResourceOptions = locale.into();
+    ///
+    /// assert_eq!(options.to_string(), "it-IT-u-ca-coptic");
+    /// assert_eq!(options.get_langid(), langid!("it-IT"));
+    /// assert_eq!(options.language(), language!("it"));
+    /// assert_eq!(options.script(), None);
+    /// assert_eq!(options.region(), Some(region!("IT")));
+    ///
+    /// let locale = options.into_locale();
+    /// assert_eq!(locale.to_string(), "it-IT-u-ca-coptic");
+    /// ```
+    pub fn into_locale(self) -> Locale {
+        let mut loc = Locale {
+            id: self.langid,
+            ..Default::default()
+        };
+        loc.extensions.unicode.keywords = self.keywords;
+        loc
+    }
+
+    /// Returns the [`Language`] for this [`ResourceOptions`].
+    pub fn language(&self) -> Language {
+        self.langid.language
+    }
+
+    /// Returns the [`Script`] for this [`ResourceOptions`].
+    pub fn script(&self) -> Option<Script> {
+        self.langid.script
+    }
+
+    /// Returns the [`Region`] for this [`ResourceOptions`].
+    pub fn region(&self) -> Option<Region> {
+        self.langid.region
+    }
+
+    /// Gets the value of the specified Unicode extension keyword for this [`ResourceOptions`].
+    pub fn get_unicode_ext(&self, key: &unicode_ext::Key) -> Option<unicode_ext::Value> {
+        self.keywords.get(key).cloned()
+    }
+
+    /// Returns whether this [`ResourceOptions`] contains a Unicode extension keyword
+    /// with the specified key and value.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use icu_provider::prelude::*;
+    /// use icu_locid::{Locale, unicode_ext_key, unicode_ext_value};
+    ///
+    /// let locale: Locale = "it-IT-u-ca-coptic".parse()
+    ///     .expect("Valid BCP-47");
+    /// let options: ResourceOptions = locale.into();
+    ///
+    /// assert_eq!(
+    ///     options.get_unicode_ext(&unicode_ext_key!("hc")),
+    ///     None
+    /// );
+    /// assert_eq!(
+    ///     options.get_unicode_ext(&unicode_ext_key!("ca")),
+    ///     Some(unicode_ext_value!("coptic"))
+    /// );
+    /// assert!(options.matches_unicode_ext(
+    ///     &unicode_ext_key!("ca"),
+    ///     &unicode_ext_value!("coptic"),
+    /// ));
+    /// ```
+    pub fn matches_unicode_ext(&self, key: &unicode_ext::Key, value: &unicode_ext::Value) -> bool {
+        self.keywords.get(key) == Some(value)
     }
 }
 
@@ -433,22 +582,22 @@ mod tests {
     use super::*;
 
     struct KeyTestCase {
-        pub resc_key: ResourceKey,
+        pub key: ResourceKey,
         pub expected: &'static str,
     }
 
     fn get_key_test_cases() -> [KeyTestCase; 3] {
         [
             KeyTestCase {
-                resc_key: resource_key!("core/cardinal@1"),
+                key: resource_key!("core/cardinal@1"),
                 expected: "core/cardinal@1",
             },
             KeyTestCase {
-                resc_key: resource_key!("core/maxlengthsubcatg@1"),
+                key: resource_key!("core/maxlengthsubcatg@1"),
                 expected: "core/maxlengthsubcatg@1",
             },
             KeyTestCase {
-                resc_key: resource_key!("core/cardinal@65535"),
+                key: resource_key!("core/cardinal@65535"),
                 expected: "core/cardinal@65535",
             },
         ]
@@ -457,39 +606,30 @@ mod tests {
     #[test]
     fn test_options_to_string() {
         for cas in get_key_test_cases().iter() {
-            assert_eq!(cas.expected, cas.resc_key.to_string());
-            writeable::assert_writeable_eq!(&cas.resc_key, cas.expected);
+            assert_eq!(cas.expected, cas.key.to_string());
+            writeable::assert_writeable_eq!(&cas.key, cas.expected);
         }
     }
 
     struct OptionsTestCase {
-        pub resc_options: ResourceOptions,
+        pub options: ResourceOptions,
         pub expected: &'static str,
     }
 
     fn get_options_test_cases() -> [OptionsTestCase; 3] {
-        use icu_locid_macros::langid;
+        use std::str::FromStr;
         [
             OptionsTestCase {
-                resc_options: ResourceOptions {
-                    variant: None,
-                    langid: Some(LanguageIdentifier::und()),
-                },
+                options: Locale::UND.into(),
                 expected: "und",
             },
             OptionsTestCase {
-                resc_options: ResourceOptions {
-                    variant: Some(Cow::Borrowed("GBP")),
-                    langid: Some(LanguageIdentifier::und()),
-                },
-                expected: "GBP/und",
+                options: Locale::from_str("und-u-cu-gbp").unwrap().into(),
+                expected: "und-u-cu-gbp",
             },
             OptionsTestCase {
-                resc_options: ResourceOptions {
-                    variant: Some(Cow::Borrowed("GBP")),
-                    langid: Some(langid!("en-ZA")),
-                },
-                expected: "GBP/en-ZA",
+                options: Locale::from_str("en-ZA-u-cu-gbp").unwrap().into(),
+                expected: "en-ZA-u-cu-gbp",
             },
         ]
     }
@@ -497,29 +637,8 @@ mod tests {
     #[test]
     fn test_key_to_string() {
         for cas in get_options_test_cases().iter() {
-            assert_eq!(cas.expected, cas.resc_options.to_string());
-            writeable::assert_writeable_eq!(&cas.resc_options, cas.expected);
-        }
-    }
-
-    #[test]
-    fn test_resource_path_to_string() {
-        for key_cas in get_key_test_cases().iter() {
-            for options_cas in get_options_test_cases().iter() {
-                let expected = if options_cas.resc_options.is_empty() {
-                    key_cas.expected.to_string()
-                } else {
-                    format!("{}/{}", key_cas.expected, options_cas.expected)
-                };
-                let resource_path = ResourcePath {
-                    key: key_cas.resc_key,
-                    // Note: once https://github.com/rust-lang/rust/pull/80470 is accepted,
-                    // we won't have to clone here.
-                    options: options_cas.resc_options.clone(),
-                };
-                assert_eq!(expected, resource_path.to_string());
-                writeable::assert_writeable_eq!(&resource_path, expected);
-            }
+            assert_eq!(cas.expected, cas.options.to_string());
+            writeable::assert_writeable_eq!(&cas.options, cas.expected);
         }
     }
 }
